```python
# 设置全部行输出
from IPython.core.interactiveshell import InteractiveShell
InteractiveShell.ast_node_interactivity = "all"
```

# 面向对象编程：定义类


```python
#定义类
class Cook:
    "这是一个厨师类的帮助文档"       #帮助文档
    name="张三"
    age="28"
    
    #用函数表示方法
    def qiecai():
        print("我会切菜")
    
    def chaocai():
        print("我会炒菜")
```


```python
#查看类的说明文档
# Cook()
# Cook.__doc__        #首位双下划线表示定义特殊的方法，一般是系统定义的名字
?Cook
# help(Cook)
```


```python
#查看类的属性
Cook.name
Cook.age
```




    '28'




```python
#增加类的属性
Cook.height="178"
Cook.height
```




    '178'




```python
#删除属性
del Cook.height
Cook.height
```


    ---------------------------------------------------------------------------

    AttributeError                            Traceback (most recent call last)

    <ipython-input-19-5d2c3703baff> in <module>
          1 #删除属性
    ----> 2 del Cook.height
          3 Cook.height
    

    AttributeError: height



```python
#修改属性
Cook.age=36
Cook.age
```




    36




```python
#调用类的方法：类名.方法名()
Cook.qiecai()
```

    我会切菜
    

# 类的实例化：对象


```python
cook1=Cook()
```


```python
cook1  
#Cook类实例出来的一个对象，存放在某个位置。只能通过对象调用对象的属性和方法
```




    <__main__.Cook at 0x2608c4316d0>




```python
cook2=Cook()
cook2
```




    <__main__.Cook at 0x2608c431940>




```python
id(cook1)
id(cook2)
```




    2613693323584




```python
cook1.age
cook2.age
```




    36






    36




```python
cook1.age=19
```


```python
cook1.age          #可以更改已经实例化后对象的属性
```




    19




```python
cook1.age
cook2.age
```




    19






    36




```python
cook1.qiecai()       #出错。因为通过实例化对象调用方法有个传参过程
```


    ---------------------------------------------------------------------------

    TypeError                                 Traceback (most recent call last)

    <ipython-input-44-faea4a25a5df> in <module>
    ----> 1 cook1.qiecai()       #出错。因为通过实例化对象调用方法有个传参过程
    

    TypeError: qiecai() takes 0 positional arguments but 1 was given



```python
#定义类
class Cook1:
    "这是一个厨师类的帮助文档"       #帮助文档
    name="张三"
    age="28"
    
    #用函数表示方法
    def qiecai(self):
        print("我会切菜")
    
    def chaocai(self):
        print("我会炒菜")           #第一个参数位置用self占位
```


```python
cook2=Cook1()
cook2.name
cook2.age
cook2.qiecai()
cook2.chaocai()
```




    '张三'






    '28'



    我会切菜
    我会炒菜
    


```python
#定义类
class Cook2:
    "这是一个厨师类的帮助文档"       #帮助文档
    name="张三"
    age="28"
    
    #用函数表示方法
    def qiecai(self):
        print("我会切菜")
        print("参数self来自于：",id(self))
    
    def chaocai(self):
        print("我会炒菜")           #第一个参数位置用self占位
```


```python
cook3=Cook2()
cook3.name
cook3.qiecai()
id(cook3)                #id(cook3)==id(self)。self参数接收的就是实例化出来的对象本身
```




    '张三'



    我会切菜
    参数self来自于： 2613693323536
    




    2613693323536



1.self参数接收的就是实例化出来的对象本身

2.通过实例化对象调用类的方法时候，内部有一个自动传参的过程，会把对象传给第一个位置的参数

3.因此，**我们在类中定义函数（方法）的时候，第一个参数必须写上，用来接收实例化对象**

# 类的属性与实例化属性


```python
class Cook3:
    #用变量表示属性
    name="张三"
    age="25岁"
    
    #特殊函数（构造函数）
    def __init__(self):  #类中定义的函数一个参数用于接收实例对象，__init__函数也不例外
        print("__init__函数在实例化对象的时候会被执行")
        
    def qie(self):
        print("我会切菜")
    
    def chao(self):
        print("我会炒菜")
```


```python
cook7=Cook3()            #_init__()函数在实例化对象的时候会自动运行，其他函数不会
```

    __init__函数在实例化对象的时候会被执行
    

**借助__init__函数设置实例属性**


```python
class Cook3:
    #用变量表示属性
    country="中国"
    
    
    #特殊函数（构造函数）
    def __init__(self,name,age):  
        self.name=name             #把def __init__(self,name,age)中的name实参赋值给self.name。 self.name就是实例属性的名字                  
        self.age=age
        print("__init__函数在实例化对象的时候会被执行")
        
    def qie(self):
        print("我会切菜")
    
    def chao(self):
        print("我会炒菜")
```


```python
cook9=Cook3()          #报错：需要写入name和age两个参数
```


    ---------------------------------------------------------------------------

    TypeError                                 Traceback (most recent call last)

    <ipython-input-76-9ab4cd905041> in <module>
    ----> 1 cook9=Cook3()
    

    TypeError: __init__() missing 2 required positional arguments: 'name' and 'age'



```python
cook9=Cook3("王爷","38岁")
```

    __init__函数在实例化对象的时候会被执行
    


```python
cook9.country
cook9.name
cook9.age
cook9.qie()
```




    '中国'






    '王爷'






    '38岁'



    我会切菜
    


```python
class Caculator():
    
    Place_of_Origin="made in china"
    
    def __init__(self,color,brand,price):
        self.color=color
        self.brand=brand
        self.price=price
    
    def sum1(self,a,b):
        s=a+b
        return s
    
    def sum2(self,a):
        s2=0
        for i in a:
            s2+=i
        return s2
    
    def sum3(self,b):
        s3=0
        for i in b:
            i3=1/i
            s3+=i3
        return s3
    
    def sum4(self,*args):
        s=0
        for i in args:
            s+=i
        return s
        
```


```python
c1=Caculator("red","hongxin","15元/个")
```


```python
c1.color
c1.brand
c1.price
```




    'red'






    'hongxin'






    '15元/个'




```python
c1.sum1(1,2)
```




    3




```python
c1.sum2([1,4,5])
```




    10




```python
c1.sum3([1,2])
```




    1.5



 **给类绑定的方法**


```python
#第一步：定义好函数。该函数的参数中，第一个参数必须是self,self参数不参加函数的运算
def sum5(self,*args):
    "功能：求一系列数的平方和"
    s=0
    for i in args:
        s+=i**2
    return s
```


```python
sum5("self",1,2,3)         #检查函数定义的是否正确
```




    14




```python
#第二步：导库
from types import MethodType  #从types模块中导入MethodType函数
```


```python
#第三步：使用MethodType函数进行绑定
#语法：类名字.方法名=MethodType(第一步中定义好的函数名，接收该方法的类的名称)
Caculator.sum5=MethodType(sum5,Caculator)   #第一个参数是需要添加的方法函数名，第二个参数是要添加的类的名称
```


```python
cal=Caculator("蓝色","卡西欧","30")
cal.brand
cal.sum4(1,2,3)
cal.sum5(1,2,3)
```




    '卡西欧'






    6






    14



 **给实例化对象绑定方法**<br>
 给某个实例化对象绑定方法，只有该实例化对象可以使用该方法，其他的实例化对象不能使用


```python
cal1=Caculator("黑色","得力","20")
cal1.sum5=MethodType(sum5,cal1)     #第一个参数写要绑定的方法名，第二个参数写要绑定的对象名
cal.sum5(1,2,5)
```




    30



# 继承和多态

继承是子类继承父类的特征和行为，本质上是类与类之间的一种关系。<br>
通过子类继承父类拥有的属性和方法，可以减少代码的冗余


```python
#定义一个关于人的类
class Person:
    """这是关于人的一个类"""
    language="Y"
    
    def __init__(self,color,race):        #实例属性：肤色和种族
        self.color=color
        self.race=race
    
    def eat(self):
        print("技能：吃东西")
```


```python
#定义一个关于人的类
class ChinesePerson:
    """这是关于中国人的一个类"""
    language="Y"
    country="中国"
    
    def __init__(self,color,race,haircolor):     #实例属性：肤色、种族和发色
        self.color=color
        self.race=race
        self.haircolor=haircolor
    
    def eat(self):
        print("技能：吃东西")
```


```python
a=ChinesePerson("黄色","汉族","黑色")
```


```python
a.color
```




    '黄色'




```python
#使用继承的语法定义中国人这个类（错误示范）
class ChinesePerson(Person):
    "这是一个用继承语法定义的中国人的类"
    country="中国"
    
    def __init__(self,color,race,haircolor):
        self.haircolor=haircolor
    
    def sowing(self):
        print("技能：播种")
```


```python
per=ChinesePerson("黄色","汉族","黑色")
```


```python
per.language
per.country
per.sowing()
per.haircolor
```

    技能：播种
    




    '黑色'




```python
per.color    #只有加上 Person.__init__(self,color,race)语句才能继承父类的属性
```


    ---------------------------------------------------------------------------

    AttributeError                            Traceback (most recent call last)

    <ipython-input-29-4a7e48d0e971> in <module>
    ----> 1 per.color
    

    AttributeError: 'ChinesePerson' object has no attribute 'color'



```python
#使用继承的语法定义中国人这个类
class ChinesePerson(Person):
    "这是一个用继承语法定义的中国人的类"
    country="中国"
    
    def __init__(self,color,race,haircolor):
        Person.__init__(self,color,race)   #继承父类的属性。相当于self.color=color self.race=race
        self.haircolor=haircolor
    
    def sowing(self):
        print("技能：播种")
```


```python
per1=ChinesePerson("黄色","苗族","黑色")
```


```python
per1.country
per1.eat()
per1.color
per1.race
per1.haircolor
```




    '中国'



    技能：吃东西
    




    '黄色'






    '苗族'






    '黑色'



**方法重写**<br>
当父类的某个方法不完全适用于子类时，可以再子类里重新改写这个方法。此时子类和父类中都有这个函数，会优选使用子类中的


```python
class Dog:
    legs="四条腿"
    
    def bark(self):
        print("汪汪")
```


```python
dog1=Dog()
dog1.bark()
```

    汪汪
    


```python
class Husky(Dog):
    def bark(self):
        print("嗷嗷")
```


```python
erha=Husky()
erha.bark()   #子类和父类有相同名字的方法时，从子类实例化出来的对象会优先使用子类的方法      
```

    嗷嗷
    

**多态**<br>
不同的对象调用相同的方法得到不同的结果<br>
实现多态的方法：继承+方法重写


```python
class Animal:
    def bark(self):
        print("*#@$!%*&")
```


```python
class Dog(Animal):
    def bark(self):
        print("汪汪")
```


```python
class Cat(Animal):
    def bark(self):
        print("喵喵")
```


```python
class Sheep(Animal):
    def bark(self):
        print("咩咩")
```


```python
dog=Dog()
cat=Cat()
sheep=Sheep()
dog.bark()
cat.bark()
sheep.bark()
```

    汪汪
    喵喵
    咩咩
    
